/* etunnel 0.0.5
 * Copyright (C) 2003-2004 Kostas Petrikas
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>

#define SHUTDOWN_CH SIGUSR1
#define max_head_size 1024

#define T_PLAIN 1
#define T_SOCKS5 2
#define T_HTTP 3

#define ALLOW 1
#define DENY 2

/*Core structure*/
typedef struct cfg_header
{
	char shost[max_head_size],rhost[max_head_size];/*addr memory*/
	int sport,lport,rport;/*ports*/
	int type;/*proxy protocol T_*/
	char *p_pass,*p_user;/*proxy user & password*/
	int chain_flag;/*proxy chain flag*/
	char *p_list;/*chain list file*/
	char *f_log;/*log file*/
	char *ip_list;/*Allow-Deny list file*/
	int INC_UDP_F,INC_TCP_F;/*local port protocol flags*/
	int OUT_UDP_F,OUT_TCP_F;/*remote port protocol flags*/
	int silent_flag;/*silent mode flag*/
	int daemon_flag;/*daemon mode flag*/
	int ad_mode;/*Allow or Deny flag*/
}cfg_header;

void dump_help(void);
void zero_cfg(struct cfg_header *);
void elog(struct cfg_header *, char *, ...);
int socket_connect_tcp(char *, int);
int socket_connect_udp(char *, int);
int read_byte(int);
int socks5_connect(int, int, char [], int);
int http_connect(int, int, char []);
long get_file_size(char *);
int shutdown_ch(void);
int shutdown_dm(void);
int handle_connection(void);
int bind_tcp_port(struct cfg_header *);
int bind_udp_port(struct cfg_header *);
int b64_ntop(u_char const *, size_t, char *, size_t);


extern int sock,a_sock,s_sock;
extern struct cfg_header config;
