/* etunnel 0.0.5
 * Copyright (C) 2003-2004 Kostas Petrikas
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "etunnel.h"                    

int check_ip(char *ip){
	FILE *fip;
	char buff[16],ch;
	int i,ii;
	long size = get_file_size(config.ip_list);
	
	fip = fopen(config.ip_list,"r");
	if(fip == NULL){
		elog(&config,"Can't open ip list!\n");
		exit(0);
	}
	
	bzero((char *)&buff,sizeof(buff));
	
	if(config.ad_mode == DENY){
		i = 0;
		while(1){
			ii = 0;
			while(1){
				if(i == size)break;
				if(ii == 16){elog(&config,"IP is to long\n");exit(0);}
				ch = getc(fip);
				if(ch == '\n'){i++;ii++;break;}
				buff[ii] = ch;
				i++;ii++;
			}
			if(!strcmp(ip,buff)){
				elog(&config,"IP is denied by rule\n");
				return 0;
			}
			bzero((char *)&buff,sizeof(buff));
			if(i == size)break;
		}
		return 1;
	}
	if(config.ad_mode == ALLOW){
		i = 0;
		while(1){
			ii = 0;
			while(1){
				if(i == size)break;
				if(ii == 16){elog(&config,"IP is to long\n");exit(0);}
				ch = getc(fip);
				if(ch == '\n'){i++;ii++;break;}
				buff[ii] = ch;
				i++;ii++;
			}
			if(!strcmp(ip,buff))	return 1;
			bzero((char *)&buff,sizeof(buff));
			if(i == size)break;
		}
		elog(&config,"IP is denied by rule\n");
		return 0;
	}
	return 1;
}

int handle_connection(void){
	struct sigaction siga;
	struct sockaddr_in c_addr;
	FILE *f_list;
	int addr_len = sizeof(c_addr);
	long size,l,ll;
	int a_pid,o_pid,len,l_port,stat;
	char buffer[1025],ch,prot_b[1025];
	
	char *prot_p,*port_p, *host_p, *return_p;
	int   prot_l, port_l, total_l;
	int n_stat, n_line;

	if( !fork() ){
		bzero(&c_addr,addr_len);
		
		
		if( config.INC_TCP_F == 1 ){
			if((a_sock = accept(sock, (struct sockaddr *)&c_addr,&sock)) < 0){
				elog(&config,"accept()\n");
				close(sock);
				exit(0);
			}
			getpeername(a_sock,(struct sockaddr *)&c_addr,&addr_len);
			elog(&config,"Incoming TCP connection from %s:%d\n",inet_ntoa(c_addr.sin_addr),
			ntohs(c_addr.sin_port));
			if(config.ad_mode != 0)
				if(!check_ip(inet_ntoa(c_addr.sin_addr))){
					close(a_sock);exit(0);
				}
		}
		else {
			a_sock = sock;
			recvfrom(a_sock,NULL,0,0,(struct sockaddr *)&c_addr,(int *)&addr_len);
			elog(&config,"Incoming UDP connection from %s:%d\n",inet_ntoa(c_addr.sin_addr),
			ntohs(c_addr.sin_port));
			if(config.ad_mode != 0)
				if(!check_ip(inet_ntoa(c_addr.sin_addr))){
					close(a_sock);exit(0);
				}
		}
		
		siga.sa_handler = shutdown_ch;
		sigemptyset(&siga.sa_mask);
		siga.sa_flags = 0;
	
		if( sigaction(SHUTDOWN_CH, &siga, NULL) ){
			elog(&config,"sigaction SHUTDOWN_CH\n");
			exit(1);
		}		
		
		if( (config.type == T_SOCKS5) && (!config.chain_flag) ){
			elog(&config,"Connecting to %s:%d ...\n",config.shost,config.sport);
			s_sock=socket_connect_tcp(config.shost,config.sport);
			if( s_sock < 0 ) exit(0);
			if( config.OUT_TCP_F == 1 ){
				if( !socks5_connect(s_sock, config.rport, config.rhost, 0)){
					close(s_sock);close(a_sock);
					exit(0);
				}
			}
			else{
				if( !socks5_connect(s_sock, config.rport, config.rhost, 1)){
					close(s_sock);close(a_sock);
					exit(0);
				}
			}
			elog(&config,"Connected!\n");
		}
		else if( config.chain_flag == 1 ){
			l = ll = prot_l = port_l = total_l = 0;
			size = get_file_size(config.p_list);
	
			if( size == -1 ){
				elog(&config,"can't open list file!\n");
				exit(0);
			}

			f_list = fopen(config.p_list,"r");
			if( f_list == NULL )	{
				elog(&config,"can't open list file!\n");
				exit(0);
			}
			n_stat=-1;
			n_line=0;
			elog(&config,"Starting chain...\n");
			while(1){
				n_line++;
				port_l=total_l=prot_l=l=0;
	
				while(1){
					if( (ll+l) >= size ) {n_stat = 0; break;}
					ch = getc(f_list);
					if( ch == '\n' ) {n_stat = 1; break;}
					switch(ch){
						case '@':{
							prot_l = l;
						}break;
						case ':':{
							port_l = l;
						}break;
					}
					l++;
				}
				if((prot_l==0)||(port_l==0)){
					elog(&config,"Syntax error in line %d\n",n_line);
					exit(0);
				}

				fseek(f_list,ll,SEEK_SET);
				total_l = l; ll += l;l = 0;
		
				prot_p = (char *)malloc(prot_l+1);
				if(prot_p == NULL){elog(&config,"memmory problem\n");exit(0);}
				bzero(prot_p,prot_l+1);
		
				return_p = prot_p;
		
				while(1){
					if(l==prot_l)break;
					ch = getc(f_list);
					*prot_p = ch;prot_p++;			
					l++;
				}
				prot_p = return_p;
		
				getc(f_list);l++;
		
				host_p = (char *)malloc((port_l+1)-prot_l);
				if(host_p == NULL){elog(&config,"memmory problem\n");exit(0);}
				bzero(host_p,(port_l+1)-prot_l);
				return_p = host_p;
		
				while(1){
					if(l==port_l)break;
					ch = getc(f_list);
					*host_p = ch;host_p++;
					l++;
				}
				host_p = return_p;
		
				getc(f_list);l++;
		
				port_p = (char *)malloc((total_l+1)-port_l);
				if(port_p == NULL){elog(&config,"memmory problem\n");exit(0);}
				bzero(port_p,(total_l+1)-port_l);
				return_p = port_p;
		
				while(1){
					if(l==total_l)break;
					ch = getc(f_list);
					*port_p = ch;port_p++;
					l++;
				}
				port_p = return_p;
				elog(&config,"Connecting to %s:%d\n",host_p,atoi(port_p));
				/***********************************************/
				if( (n_line == 1) && (!strcmp("socks5",prot_p) || !strcmp("http",prot_b) )){
					s_sock=socket_connect_tcp(host_p,atoi(port_p));
					if( s_sock < 0 )
						{exit(0);}
				}
				else if( !strcmp("socks5",prot_b) ){	
					if( !socks5_connect(s_sock,atoi(port_p),host_p,0) ){
						close(s_sock);close(a_sock);
						exit(0);
					}
				}
				else if( !strcmp("http",prot_b) ){	
					if( !http_connect(s_sock,atoi(port_p),host_p) ){
						close(s_sock);close(a_sock);
						exit(0);
					}
				}
				else{	
					elog(&config,"Unknown proxy protocol! Skipping...\n");
				}
				bzero((char *)&prot_b,sizeof(prot_b));
				strncpy(prot_b,prot_p,sizeof(prot_b)-1);
				/***********************************************/
				free(host_p);free(port_p);
				if(!n_stat)break;
				else getc(f_list);ll++;
			}
			fclose(f_list);
			if( !strcmp("socks5",prot_p) ){
				if( config.OUT_TCP_F == 1 ){
					if( !socks5_connect(s_sock,config.rport,config.rhost,0)){
						close(s_sock);close(a_sock);exit(0);
					}
				}
				else{
					if( !socks5_connect(s_sock,config.rport,config.rhost,1)){
						close(s_sock);close(a_sock);exit(0);
					}
				}
			}
			else if( !strcmp("http",prot_p) && config.OUT_TCP_F == 1 ){
				if( !http_connect(s_sock,config.rport,config.rhost)){
					close(s_sock);close(a_sock);
					exit(0);
				}
			}
			else if( !strcmp("http",prot_p) && config.OUT_UDP_F == 1 ){
				elog(&config,"There is no UDP support via HTTP proxy\n");
				close(s_sock);close(a_sock);exit(0);
			}
			else {
				elog(&config,"Unknown proxy protocol! Closing...\n");
				close(s_sock);close(a_sock);exit(0);
			}
			elog(&config,"Proxy chain completed. Tunnel opened!\n");
			free(prot_p);
		}
		else if( config.type == T_PLAIN ){
			if( config.OUT_TCP_F == 1 ){
				elog(&config,"Connecting to TCP %s:%d ...\n",config.rhost,config.rport);
				s_sock=socket_connect_tcp(config.rhost,config.rport);
			}
			else{
				elog(&config,"Connecting to UDP %s:%d ...\n",config.rhost,config.rport);
				s_sock=socket_connect_udp(config.rhost,config.rport);
			}
			if( s_sock < 0 ) exit(0);
			elog(&config,"Connected!\n");
		}
		else if( (config.type == T_HTTP) && (!config.chain_flag) ){
			elog(&config,"Connecting to %s:%d ...\n",config.shost,config.sport);
			s_sock=socket_connect_tcp(config.shost,config.sport);
			if( s_sock < 0 ) exit(0);
			if(!http_connect(s_sock, config.rport, config.rhost)){
				elog(&config,"closing...\n");
				close(s_sock);close(a_sock);
				exit(0);
			}
			elog(&config,"Connected!\n");
		}
		
		a_pid=getpid();
		
		if( !(o_pid=fork()) ){
			bzero((char *)&buffer,sizeof(buffer));
				while( (len = recv(a_sock,buffer,(sizeof(buffer)-1),0)) > 0 ){
					if( send(s_sock,buffer,len,0) < 0 ){
						close(s_sock);
						bzero((char *)&buffer,sizeof(buffer));
						elog(&config,"Socket Killed (Server output)\n");
						kill(a_pid,SHUTDOWN_CH);
						exit(0);
					}
					bzero((char *)&buffer,sizeof(buffer));
				}
			elog(&config,"Socket Killed (Server input)\n");
			kill(a_pid,SIGKILL);
			exit(0);
		}
		else{
			bzero((char *)&buffer,sizeof(buffer));
			while( (len = recv(s_sock,buffer,(sizeof(buffer)-1),0)) > 0 ){
				if( config.INC_TCP_F == 1 ){
					if( send(a_sock,buffer,len,0) < 0 ){
						close(a_sock);
						bzero((char *)&buffer,sizeof(buffer));
						elog(&config,"Socket Killed (Client TCP output)\n");
						kill(o_pid,SIGKILL);
						exit(0);
					}
				}
				else{
					if( sendto(a_sock,buffer,len,0,&c_addr,addr_len) < 0 ){
						close(a_sock);
						bzero((char *)&buffer,sizeof(buffer));
						elog(&config,"Socket Killed (Client UDP output)\n");
						kill(o_pid,SIGKILL);
						exit(0);
					}
				}
				bzero((char *)&buffer,sizeof(buffer));
			}
			elog(&config,"Socket Killed (Client input)\n");
			kill(o_pid,SIGKILL);
			exit(0);			
		}
		exit(0);
	}
	return 0;
}
