/* etunnel 0.0.5
 * Copyright (C) 2003-2004 Kostas Petrikas
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*     POSIX4 93/09        */ 
#define POSIX_C_SOURCE 199309

#define len strlen(argv[stat+1])+1

#include "etunnel.h"

int sock,a_sock,s_sock;
struct cfg_header config;

void dump_help(void){
	printf("etunnel [options] [local port] [port protocol] [remote host] [remote port] [port protocol]\n");
	printf("\nOptions: -t [type] -H [host] -P [port] -c [list_file] -l [log_file] -u [user] -p [pass] -D [list_file] -A [list_file] -s\n");
}

void zero_cfg(struct cfg_header *cfg){
	bzero(cfg->shost,max_head_size);
	bzero(cfg->rhost,max_head_size);
	cfg->sport=cfg->lport=cfg->rport=
	cfg->type=cfg->chain_flag=0;
	cfg->silent_flag=cfg->daemon_flag=0;
	cfg->INC_UDP_F=cfg->INC_TCP_F=0;
	cfg->OUT_UDP_F=cfg->OUT_TCP_F=0;
	cfg->ad_mode=0;
	cfg->f_log=cfg->ip_list=NULL;
}

long get_file_size(char *path){
	FILE *ffile;
	long size;
	
	ffile = fopen(path,"r");
	if( ffile == NULL ) return -1;

	fseek(ffile, 0, 2);
	size=  ftell(ffile);
	fclose(ffile);
	
	return size;
}

int shutdown_ch(void){
	close(a_sock);
	close(s_sock);
	exit(0);
	return 0;
}

int shutdown_dm(void){
	printf("\nGot SIGINT, shutting down server...\n");

	printf("Killing connections...\n");
	kill(0, SHUTDOWN_CH);

	exit(0);
	
	return 0;
}

void elog(struct cfg_header *cfg,char *fmt,...){
	va_list args;
	FILE *flog;
	time_t ti;
	char *chtime;

	(void)time(&ti);
	chtime = ctime(&ti);
	chtime[strlen(chtime)-1] = 0;
	
	if(!cfg->silent_flag){
		if(cfg->f_log != NULL){
			flog = fopen(cfg->f_log,"a");
			if(flog != NULL){
				va_start(args,fmt);
					printf("%s :: ",chtime);
					vfprintf(flog,fmt,args);
				va_end(args);
				fclose((FILE *)cfg->f_log);
			}
		}
		else if(!cfg->daemon_flag){
			va_start(args,fmt);
				printf("%s :: ",chtime);
				vprintf(fmt,args);
			va_end(args);
		}
	}
}

int main(int argc, char **argv){
	struct sigaction siga;
	int dpid;
	int arg = (argc-1),stat = 0,argg = (arg-5);
	
	zero_cfg(&config);

	if ( arg < 5 ) { dump_help(); return 0; }
	
	while( stat != (argg+1) ){
		if( (strlen(argv[stat]) == 2) && (argv[stat][0]=='-') ){
			switch(argv[stat][1]){
				case 'H':{
					if( (argg<(stat+1)) || (argv[stat+1][0]=='-') ){
						printf("-H Nead value\n");dump_help();
						return 0;}
					strncpy(config.shost,argv[stat+1],max_head_size);
					stat++;
				}break;
				case 'P':{
					if( (argg<(stat+1)) || (argv[stat+1][0]=='-') ){
						printf("-P Nead value\n");dump_help();
						return 0;}
					config.sport = atoi(argv[stat+1]);
					if(!config.sport){
						printf("-P Bad value\n");dump_help();
						return 0;
					}
					stat++;
				}break;
				case 's':{
					config.silent_flag = 1;
				}break;
				case 'd':{
					config.daemon_flag = 1;
				}break;
				case 't':{
					if( (argg<(stat+1)) || (argv[stat+1][0]=='-') ){
						printf("-t Nead value\n");dump_help();
						return 0;}
					if( !strcmp(argv[stat+1],"plain") ){
						config.type = T_PLAIN;
					}
					else if( !strcmp(argv[stat+1],"socks5") ){
						config.type = T_SOCKS5;
					}
					else if( !strcmp(argv[stat+1],"http") ){
						config.type = T_HTTP;
					}
					else{
						printf("-t Bad value\n");dump_help();
						return 0;
					}
					stat++;
				}break;
				case 'u':{
					if( (argg<(stat+1)) || (argv[stat+1][0]=='-') ){
						printf("-u Nead value\n");dump_help();
						return 0;}
					config.p_user = (char *)malloc(len);
					if(config.p_user == NULL)return 0;
					bzero(config.p_user,len);
					strncpy(config.p_user,argv[stat+1],len-1);
					stat++;
				}break;
				case 'p':{
					if( (argg<(stat+1)) || (argv[stat+1][0]=='-') ){
						printf("-p Nead value\n");dump_help();
						return 0;}
					config.p_pass = (char *)malloc(len);
					if(config.p_pass == NULL)return 0;
					bzero(config.p_pass,len);
					strncpy(config.p_pass,argv[stat+1],len-1);
					stat++;
				}break;
				case 'A':{
					if( (argg<(stat+1)) || (argv[stat+1][0]=='-') ){
						printf("-A Nead value\n");dump_help();
						return 0;}
					config.ad_mode = ALLOW;
					config.ip_list = (char *)malloc(len);
					if(config.ip_list == NULL)return 0;
					bzero(config.ip_list,len);
					strncpy(config.ip_list,argv[stat+1],len-1);
					stat++;
				}break;
				case 'D':{
					if( (argg<(stat+1)) || (argv[stat+1][0]=='-') ){
						printf("-D Nead value\n");dump_help();
						return 0;}
					config.ad_mode = DENY;
					config.ip_list = (char *)malloc(len);
					if(config.ip_list == NULL)return 0;
					bzero(config.ip_list,len);
					strncpy(config.ip_list,argv[stat+1],len-1);
					stat++;
				}break;
				case 'l':{
					if( (argg<(stat+1)) || (argv[stat+1][0]=='-') ){
						printf("-l Nead value\n");dump_help();
						return 0;}

					config.f_log = (char *)malloc(len);
					bzero(config.f_log,len);
					strncpy(config.f_log,argv[stat+1],len-1);
					stat++;
				}break;
				case 'c':{
					if( (argg<(stat+1)) || (argv[stat+1][0]=='-') ){
						printf("-c Nead value\n");dump_help();
						return 0;}

					config.chain_flag = 1;
					config.p_list = (char *)malloc(len);
					bzero(config.p_list,len);
					strncpy(config.p_list,argv[stat+1],len-1);
					stat++;
				}break;
				default :{
					printf("Unknown option %s\n",argv[stat]);
					dump_help();
					return 0;
				}break;
			}
		}
		stat++;
	}
	if( !(config.lport = atoi(argv[stat])) ){
		printf("Bad local port value\n");dump_help();
		return 0;}
	stat++;
	if( !strcmp(argv[stat],"UDP") ){
		config.INC_UDP_F = 1;
	}
	else if( !strcmp(argv[stat],"TCP") ){
		config.INC_TCP_F = 1;
	}
	else {printf("Local port protocol invalid\n");exit(0);}
	stat++;strncpy(config.rhost,argv[stat],max_head_size);
	stat++;if( !(config.rport = atoi(argv[stat])) ){
		printf("Bad remote port value\n");dump_help();
		return 0;}
	stat++;
	if( !strcmp(argv[stat],"UDP") ){
		config.OUT_UDP_F = 1;
	}
	else if( !strcmp(argv[stat],"TCP") ){
		config.OUT_TCP_F = 1;
	}
	else {printf("Remote port protocol invalid %s\n",argv[stat]);exit(0);}
	
	if( !config.type )config.type = T_PLAIN;
	else if( (config.type == T_SOCKS5) && (!config.chain_flag) ){
		if(!strlen(config.shost)){printf("Nead option -H\n");return 0;}
	}
	if( strlen(config.shost) > 0){
		if( !config.sport ){
			printf("Nead option -P\n");
			return 0;
		}
	}
	if( (config.type == T_HTTP) && (config.OUT_UDP_F == 1) ){
		printf("There is no UDP support via HTTP proxy\n");
		return 0;
	}
	
	if( config.daemon_flag == 1 ){
		if( (dpid = fork()) == -1){
			printf("Can not fork daemon\n");
			return 0;
		}
		else if (dpid){
			printf("Daemon started\n");
			return 0;
		}
		setsid();
		chdir("/");
	}

	siga.sa_handler = shutdown_dm;
	sigemptyset(&siga.sa_mask);
	siga.sa_flags = 0;
	
	if( sigaction(SIGINT, &siga, NULL) ){
		printf("sigaction SIGINT\n");
		exit(1);
	}
	
	if( config.INC_UDP_F == 1 ) bind_udp_port(&config);
	else bind_tcp_port(&config);

	while(1)pause();
	
	return 0;
}
