/* etunnel 0.0.5
 * Copyright (C) 2003-2004 Kostas Petrikas
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "etunnel.h"

int socket_connect_tcp(char *hostname,int portnum){
	struct sockaddr_in sa;
	struct hostent *hp;
	int s;

	if((hp= gethostbyname(hostname)) == NULL) {
		elog(&config,"Requesting hostname failed\n");
		return -1;
	}

	bzero(&sa,sizeof(sa));
	bcopy(hp->h_addr,(char *)&sa.sin_addr,hp->h_length);
	sa.sin_family = hp->h_addrtype;
	sa.sin_port = htons((u_short)portnum);

	if((s=socket(hp->h_addrtype,SOCK_STREAM,0)) < 0){
		elog(&config,"Socket creation failed %s:%d\n",hostname,portnum);
		return -1;
	}
	if (connect(s,&sa,sizeof sa) < 0){
		close(s);
		elog(&config,"TCP connection failed\n");
		return -1;
	}
	return(s);
}

int socket_connect_udp(char *hostname,int portnum){
	struct sockaddr_in sa;
	struct hostent *hp;
	int s;

	if((hp= gethostbyname(hostname)) == NULL) {
		elog(&config,"Requesting hostname failed\n");
		return -1;
	}

	bzero(&sa,sizeof(sa));
	bcopy(hp->h_addr,(char *)&sa.sin_addr,hp->h_length);
	sa.sin_family = hp->h_addrtype;
	sa.sin_port = htons((u_short)portnum);

	if((s=socket(hp->h_addrtype,SOCK_DGRAM,IPPROTO_UDP)) < 0){
		("Socket creation failed %s:%d\n",hostname,portnum);
		return -1;
	}
	if (connect(s,&sa,sizeof sa) < 0){
		close(s);
		elog(&config,"UDP connection failed\n");
		return -1;
	}
	return(s);
}

int read_byte(int fd){
	unsigned char c;

	if(read(fd, &c, 1) != -1)
		return c;

	return -1;
}

int bind_tcp_port(struct cfg_header *config){
	struct sockaddr_in addr_;
	struct sigaction siga;
	
	if( (sock = socket(AF_INET,SOCK_STREAM,0)) < 0 ){
		elog(config,"Failed to create socket\n");
		exit(0);
	}

	bzero((void *)&addr_,sizeof(addr_));

	addr_.sin_family = AF_INET;
	addr_.sin_port = htons(config->lport);
	addr_.sin_addr.s_addr = htons(INADDR_ANY);

	if( bind(sock, (struct sockaddr *)&addr_, sizeof(addr_)) == -1 ){
		elog(config,"Failed to bind port %d\n",config->lport);
		exit(0);
	}
	
	if( listen(sock, 5) < 0 )elog(config,"listen()\n");
	
	siga.sa_handler = handle_connection;
	sigemptyset(&siga.sa_mask);
	siga.sa_flags = 0;
	
	if( sigaction(SIGIO, &siga, NULL) ){
		elog(config,"sigaction SIGIO\n");
		exit(1);
	}	
	if ( fcntl(sock, F_SETOWN, getpid() ) < 0) {
		elog(config,"fcntl F_SETOWN\n");
		exit(1);
	}

	if ( fcntl(sock, F_SETFL, FASYNC) < 0 ) {
		elog(config,"fcntl F_SETFL, FASYNC\n");
		exit(1);
	}
	return 0;
}

int bind_udp_port(struct cfg_header *config){
	struct sockaddr_in addr_;
	struct sigaction siga;	
	
	if( (sock = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP)) < 0 ){
		elog(config,"Failed to create socket\n");
		exit(0);
	}
	bzero((char *)&addr_,sizeof(addr_));

	addr_.sin_family = AF_INET;
	addr_.sin_port = htons(config->lport);
	addr_.sin_addr.s_addr = htons(INADDR_ANY);

	if( bind(sock, (struct sockaddr *)&addr_, sizeof(addr_)) == -1 ){
		elog(config,"Failed to bind port %d\n",config->lport);
		exit(0);
	}
	
	siga.sa_handler = handle_connection;
	sigemptyset(&siga.sa_mask);
	siga.sa_flags = 0;
	
	if( sigaction(SIGIO, &siga, NULL) ){
		elog(config,"sigaction SIGIO\n");
		exit(1);
	}	
	
	if ( fcntl(sock, F_SETOWN, getpid() ) < 0) {
		elog(config,"fcntl F_SETOWN\n");
		exit(1);
	}

	if ( fcntl(sock, F_SETFL, FASYNC) < 0 ) {
		elog(config,"fcntl F_SETFL, FASYNC\n");
		exit(1);
	}
	return 0;
}
