/* etunnel 0.0.5
 * Copyright (C) 2003-2004 Kostas Petrikas
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "etunnel.h"

int socks5_connect(int sck, int rport, char rhost[], int udp_f){
	char establish_o[2];
	char request_o[3],*login_o;
	int byte,b,len,len_u,len_p;

	char *buff,*buff_p;

	establish_o[0] = 0x05;
	establish_o[1] = 0x01;
	if((config.p_user == NULL)||(config.p_pass == NULL))
		establish_o[2] = 0x00;
	else
		establish_o[2] = 0x02;

	send(sck,establish_o,sizeof(establish_o),0);
	if((b = read_byte(sck)) != 0x05){
		elog(&config,"SOCKS5 -> Bad version replie 0x%x\n",b);
		return 0;
	}
	if((b = read_byte(sck)) != establish_o[2]){
		if(b!=0x00){
			elog(&config,"SOCKS5 -> Bad replie 0x%x\n",b);
			return 0;
		}
	}
	
	else if((config.p_user != NULL)&&(config.p_pass != NULL)){
		len_u = strlen(config.p_user);
		len_p = strlen(config.p_pass);
		login_o = (void *)malloc(3+len_u+len_p);
		if(!login_o)return 0;
		login_o[0] = 0x01;
		login_o[1] = len_u;
		memcpy(login_o+2,config.p_user,len_u);
		login_o[2+len_u] = len_p;
		memcpy(login_o+3+len_u,config.p_pass,len_p);
		send(sck,login_o,3+len_u+len_p,0);
		
		if(read_byte(sck) != 0x01 || read_byte(sck) != 0x00) {
			elog(&config,"SOCKS5 -> Authentication failed\n");
			close(sck);
			return 0;
		}
		free(login_o);
	}
	
	request_o[0]    = 0x05;
	if(!udp_f)
		request_o[1] = 0x01;
	else
		request_o[1] = 0x03;
	request_o[2]    = 0x00;
	request_o[3]    = 0x03;
	
	buff=(char *)malloc(7+strlen(rhost));
	bzero(buff,7+strlen(rhost));

	memcpy(buff,&request_o,4);
	buff_p=buff+4;
	*buff_p=strlen(rhost);
	memcpy((buff+5),rhost,strlen(rhost));
	buff_p=buff+(5+strlen(rhost));
	*buff_p=(rport >> 8);buff_p++;
	*buff_p=(rport & 0xff);

	send(sck,buff,7+strlen(rhost),0);
	
	if(read_byte(sck)!=0x05){
		elog(&config,"SOCKS5 -> Bad version replie\n");
		return 0;
	}

	if( (byte = read_byte(sck)) != (char)0x00 ){
		switch(byte){
			case 0x01:{
				elog(&config,"SOCKS5 -> General SOCKS server failure\n");
				return 0;
			}break;
			case 0x02:{
				elog(&config,"SOCKS5 -> Connection not allowed by ruleset\n");
				return 0;
			}break;
			case 0x03:{
				elog(&config,"SOCKS5 -> Network unreachable\n");
				return 0;
			}break;
			case 0x04:{
				elog(&config,"SOCKS5 -> Host unreachable\n");
				return 0;
			}break;
			case 0x05:{
				elog(&config,"SOCKS5 -> Connection refused\n");
				return 0;
			}break;
			case 0x06:{
				elog(&config,"SOCKS5 -> TTL expired\n");
				return 0;
			}break;
			case 0x07:{
				elog(&config,"SOCKS5 -> Command not supported\n");
				return 0;
			}break;
			case 0x08:{
				elog(&config,"SOCKS5 -> Address type not supported\n");
				return 0;
			}break;
			default :{
				elog(&config,"SOCKS5 -> Bad replie\n");
				return 0;
			}break;				
		}
	}
	read_byte(sck);
	byte = read_byte(sck);
	if(byte == 0x01){
		for(b=0; b<4; b++){
			read_byte(sck);
		}
	}
	else if(byte == 0x03){
		len = read_byte(sck);
		for(b=0; b<len; b++){
			read_byte(sck);
		}
	}
	else{
		elog(&config,"SOCKS5 -> Bad replie\n");
		return 0;
	}
	for(b=0; b<2; b++){
		read_byte(sck);
	}
	
	elog(&config,"SOCKS5 -> Connected!\n");
	
	return 1;
}
